/*
 * com.brian_web.x509.* - By Brian Alliet
 * Copyright (C) 2004 Brian Alliet
 * 
 * Based on Bouncy Castle by The Legion Of The Bouncy Castle
 * Copyright (c) 2000 The Legion Of The Bouncy Castle 
 * (http://www.bouncycastle.org)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDER.S BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package org.ibex.crypto;

import java.io.*;
import java.util.*;

public class X509 {
    private X509() { }
    
    public static class Certificate {
        public static final String RSA_ENCRYPTION           = "1.2.840.113549.1.1.1";
        public static final String MD2_WITH_RSA_ENCRYPTION  = "1.2.840.113549.1.1.2";
        public static final String MD4_WITH_RSA_ENCRYPTION  = "1.2.840.113549.1.1.3";
        public static final String MD5_WITH_RSA_ENCRYPTION  = "1.2.840.113549.1.1.4";
        public static final String SHA1_WITH_RSA_ENCRYPTION = "1.2.840.113549.1.1.5";
    
        public static final String BASIC_CONSTRAINTS = "2.5.29.19";
    
        private final byte[] certBytes;
        private final byte[] tbsCertBytes;
    
        public final Number version;    
        public final Number serialNo;
        public final X509.Name issuer;
        public final Date startDate;
        public final Date endDate;
        public final X509.Name subject;
    
        public final AlgorithmIdentifier publicKeyAlgorithm;
        public final DER.BitString publicKey;
    
        public final Object issuerUniqueID;
        public final Object subjectUniqueID;
    
        public final Vector extensions;
    
        public final DER.BitString signature;
        public final AlgorithmIdentifier signatureAlgorithm;
    
        public final BC basicContraints;

    
        public Certificate(InputStream is) throws IOException {
            int i;
            RecordingInputStream certIS = new RecordingInputStream(is);
            DER.InputStream certSequence = DER.InputStream.New(certIS).getSequenceStream();
            RecordingInputStream tbsCertIS = new RecordingInputStream(certSequence);
        
            try {
                Vector tbsSequence = (Vector) DER.InputStream.New(tbsCertIS).readObject();
                tbsCertBytes = tbsCertIS.getBytes();
                signatureAlgorithm = new AlgorithmIdentifier(certSequence.readObject());
                signature = (DER.BitString) certSequence.readObject();
            
                i=0;
                if(tbsSequence.elementAt(i) instanceof DER.TaggedObject)
                    version = (Number)((DER.TaggedObject)tbsSequence.elementAt(i++)).object;
                else
                    version = new Integer(0);
            
                serialNo = (Number) tbsSequence.elementAt(i++);
                AlgorithmIdentifier signatureAlgorithm2 = new AlgorithmIdentifier(tbsSequence.elementAt(i++));
                if(!signatureAlgorithm2.equals(signatureAlgorithm))
                    throw new DER.Exception("AlgoritmIdentifier mismatch " + signatureAlgorithm + " vs " + signatureAlgorithm2);
                issuer = new X509.Name(tbsSequence.elementAt(i++));
            
                Vector validity = (Vector) tbsSequence.elementAt(i++);
                startDate = (Date) validity.elementAt(0);
                endDate = (Date) validity.elementAt(1);
            
                subject = new X509.Name(tbsSequence.elementAt(i++));
            
                Vector publicKeyInfo = (Vector) tbsSequence.elementAt(i++);
                publicKeyAlgorithm = new AlgorithmIdentifier(publicKeyInfo.elementAt(0));
                publicKey = (DER.BitString) publicKeyInfo.elementAt(1);
          
                Object issuerUniqueID_=null,subjectUniqueID_=null;
                Vector extensions_=null;
                for(;i < tbsSequence.size();i++) {
                    DER.TaggedObject to = (DER.TaggedObject) tbsSequence.elementAt(i);
                    switch(to.tag) {
                        case 1: issuerUniqueID_ = to.object; break;
                        case 2: subjectUniqueID_ = to.object; break;
                        case 3: extensions_ = (Vector) to.object; break;
                    }
                }
                issuerUniqueID = issuerUniqueID_;
                subjectUniqueID = subjectUniqueID_;
                extensions = extensions_;
            
                BC bc = null;
            
                if(extensions != null) {
                    for(Enumeration e = extensions.elements(); e.hasMoreElements(); ) {
                        Vector extension = (Vector) e.nextElement();
                        String oid = (String) extension.elementAt(0);
                        byte[] data = (byte[]) extension.elementAt(extension.size()-1);
                        if(oid.equals(BASIC_CONSTRAINTS))
                            bc = new BC(DER.InputStream.New(new ByteArrayInputStream(data)).readObject());
                    }
                }
                basicContraints = bc;
            } catch(RuntimeException e) {
                e.printStackTrace();
                throw new DER.Exception("Invalid x509 Certificate");
            }
            certBytes = certIS.getBytes();
        }
    
    
        public String getSubjectField(String fieldID) { return subject.get(fieldID); }
        public String getCN() { return getSubjectField(X509.Name.CN); }
    
        public boolean isValid() {
            Date now = new Date();
            return !now.after(endDate) && !now.before(startDate);
        }
    
        public RSA.PublicKey getRSAPublicKey() throws DER.Exception {
            if(!RSA_ENCRYPTION.equals(publicKeyAlgorithm.id)) throw new DER.Exception("This isn't an RSA public key");
            try {
                return new RSA.PublicKey(DER.InputStream.New(new ByteArrayInputStream(publicKey.data)).readObject());
            } catch(IOException e) {
                throw new DER.Exception(e.getMessage());
            } catch(RuntimeException e) {
                throw new DER.Exception("Invalid RSA Public Key " + e.getMessage());
            }
        }
    
        public boolean isSignedBy(Certificate signer) throws DER.Exception {
            return isSignedWith(signer.getRSAPublicKey());
        }
        public boolean isSignedWith(RSA.PublicKey rsapk) throws DER.Exception {
            try {
                Digest digest;
                if(signatureAlgorithm.id.equals(MD5_WITH_RSA_ENCRYPTION)) digest = new MD5();
                else if(signatureAlgorithm.id.equals(SHA1_WITH_RSA_ENCRYPTION)) digest = new SHA1();
                else if(signatureAlgorithm.id.equals(MD2_WITH_RSA_ENCRYPTION)) digest = new MD2();
                else throw new DER.Exception("Unknown signing algorithm: " + signatureAlgorithm.id);
                        
                PKCS1 pkcs1 = new PKCS1(new RSA(rsapk.modulus,rsapk.exponent,true));
                byte[] d = pkcs1.decode(signature.data);
            
                Vector v = (Vector) DER.InputStream.New(new ByteArrayInputStream(d)).readObject();
                byte[] signedDigest = (byte[]) v.elementAt(1);
                            
                if(signedDigest.length != digest.getDigestSize()) return false;
            
                digest.update(tbsCertBytes,0,tbsCertBytes.length);
                byte[] ourDigest = new byte[digest.getDigestSize()];
                digest.doFinal(ourDigest,0);
            
                for(int i=0;i<digest.getDigestSize();i++) if(ourDigest[i] != signedDigest[i]) return false;
                return true;
            }
            catch(RuntimeException e) { e.printStackTrace(); return false; }
            catch(PKCS1.Exn e) { e.printStackTrace(); return false; }
            catch(IOException e) { e.printStackTrace(); return false; }
        }
    
        public byte[] getMD5Fingerprint() { return getFingerprint(new MD5()); }
        public byte[] getSHA1Fingerprint() { return getFingerprint(new SHA1()); }
        public byte[] getFingerprint(Digest h) {
            h.update(certBytes,0,certBytes.length);
            byte[] digest = new byte[h.getDigestSize()];
            h.doFinal(digest,0);
            return digest;
        }
    
        private class RecordingInputStream extends FilterInputStream {
            public ByteArrayOutputStream baos = new ByteArrayOutputStream();
            private boolean on = true;
            public void on() { on = true; }
            public void off() { on = false; }
            public RecordingInputStream(InputStream is) { super(is); }
            public int read() throws IOException {
                int n = super.read();
                if(n != -1 && on) baos.write(n);
                return n;
            }
            public int read(byte[] buf, int off, int len) throws IOException {
                int n = super.read(buf,off,len);
                if(n != -1 && on) baos.write(buf,off,n);
                return n;
            }
            public byte[] getBytes() { return baos.toByteArray(); }
        }
    
        public static class BC {
            public final boolean isCA;
            public final Number pathLenConstraint;
            BC(Object o) {
                Vector seq = (Vector) o;
                isCA = seq.size() > 0 ? ((Boolean) seq.elementAt(0)).booleanValue() : false;
                pathLenConstraint = seq.size() > 1 ? (Number) seq.elementAt(1) : null;
            }
        }
    
        public static class AlgorithmIdentifier {
            public final String id;
            public final Object parameters;
        
            AlgorithmIdentifier(Object o) {
                Vector seq = (Vector) o;
                id = (String) seq.elementAt(0);
                parameters = seq.elementAt(1);
            }
            public boolean equals(Object o_) {
                if(o_ == this) return true;
                if(!(o_ instanceof AlgorithmIdentifier)) return false;
                AlgorithmIdentifier o = (AlgorithmIdentifier) o_;
                return o.id.equals(id) && o.parameters.equals(parameters);
            }
            public int hashCode() { return id.hashCode() ^ parameters.hashCode(); }
        }
    
        /*public static void main(String[] args) throws Exception {
          Certificate cert = new Certificate(new FileInputStream(args[0]));
          System.err.println("CN: " + cert.getCN());
          System.err.println("Subject: " + cert.subject);
          System.err.println("Issuer: " + cert.issuer);
          System.err.println("Start Date: " + cert.startDate);
          System.err.println("End Date: " + cert.endDate);
          System.err.println("SHA1 Fingerprint: " + prettyBytes(cert.getSHA1Fingerprint()));
          RSA.PublicKey key = cert.getRSA.PublicKey();
          System.err.println("Modulus: " + prettyBytes(key.modulus.toByteArray()));
          System.err.println("Exponent: " + key.exponent);
          System.err.println("Signature: " + prettyBytes(cert.signature.data));
          }
    
          public static String prettyBytes(byte[] fp) {
          StringBuffer sb = new StringBuffer(fp.length*3);
          for(int i=0;i<fp.length;i++) {
          if(i>0) sb.append(":");
          sb.append("0123456789abcdef".charAt((fp[i] & 0xf0) >>> 4));
          sb.append("0123456789abcdef".charAt((fp[i] & 0x0f) >>> 0));
          }
          return sb.toString();
          }*/
    }

    public static class Name {
        // Some common OIDs
        public static final String C = "2.5.4.6";
        public static final String O = "2.5.4.10";
        public static final String T = "2.5.4.12";
        public static final String SN = "2.5.4.5";
        public static final String L = "2.5.4.7";
        public static final String ST = "2.5.4.8";
        public static final String OU = "2.5.4.11";
        public static final String CN = "2.5.4.3";
        public static final String E = "1.2.840.113549.1.9.1";
    
        private final Vector keys = new Vector();
        private final Vector values = new Vector();
    
        public Name(Object seq_) throws DER.Exception {
            try {
                Vector seq = (Vector) seq_;
                for(Enumeration e = seq.elements();e.hasMoreElements();) {
                    Vector component = (Vector) ((Vector)e.nextElement()).elementAt(0);
                    keys.addElement(component.elementAt(0));
                    values.addElement(component.elementAt(1));
                }
            } catch(RuntimeException e) {
                e.printStackTrace();
                throw new DER.Exception("Invalid Name " + e.toString());
            }
        }
    
        public boolean equals(Object o_) {
            if(o_ instanceof String) return toString().equals(o_);
            if(!(o_ instanceof Name)) return false;
            Name o = (Name) o_;
            if(keys.size() != o.keys.size()) return false;
            int size = keys.size();
            for(int i=0;i<size;i++) {
                String oid = (String) keys.elementAt(i);
                String oid2 = (String) o.keys.elementAt(i);
                if(!oid.equals(oid2)) return false;
            
                String val1 = (String) values.elementAt(i);
                String val2 = (String) o.values.elementAt(i);
                if(val1.equals(val2)) continue;
            
                val1 = val1.trim().toLowerCase();
                val2 = val2.trim().toLowerCase();
                if(val1.equals(val2)) continue;
            
                val1 = removeExtraSpaces(val1);
                val2 = removeExtraSpaces(val2);
                if(val1.equals(val2)) continue;
            
                return false;
            }
            return true;
        }
    
        public int hashCode() { return keys.hashCode() ^ values.hashCode(); }
    
        public String get(String fieldID) {
            int i = keys.indexOf(fieldID);
            return i == -1 ? null : (String)values.elementAt(i);
        }
    
        public String[] getOIDs() {
            String[] ret = new String[keys.size()];
            keys.copyInto(ret);
            return ret;
        }
    
        public String[] getValues() {
            String[] ret = new String[values.size()];
            values.copyInto(ret);
            return ret;
        }
    
        private static String removeExtraSpaces(String s) {
            if(s.indexOf(' ') == -1) return s;
            StringBuffer sb = new StringBuffer(s.length());
            int l = s.length();
            boolean inWhitespace = false;
            for(int i=0;i<l;i++) {
                if(s.charAt(i) == ' ') {
                    if(inWhitespace) continue;
                    inWhitespace = true;
                } else if(inWhitespace) {
                    inWhitespace = false;
                }
                sb.append(s.charAt(i));
            }
            return sb.toString();
        }
    
        private final static Hashtable oidMap = new Hashtable();
        static {
            oidMap.put(Name.C,"C");
            oidMap.put(Name.O,"O");
            oidMap.put(Name.T,"T");
            oidMap.put(Name.SN,"SN");
            oidMap.put(Name.L,"L");
            oidMap.put(Name.ST,"ST");
            oidMap.put(Name.OU,"OU");
            oidMap.put(Name.CN,"CN");
            oidMap.put(Name.E,"E");
        }
    
        public String toString() {
            StringBuffer sb = new StringBuffer();
            int size = keys.size();
            for(int i=0;i<size;i++) {
                if(sb.length() > 0) sb.append(",");
                String fieldID = (String) keys.elementAt(i);
                String fieldName = (String) oidMap.get(fieldID);
                sb.append(fieldName != null ? fieldName : fieldID).append("=").append(values.elementAt(i));
            }
            return sb.toString();
        }
    }
}
